/*
 * flimsy_dates.cxx - date and time manipulation functions
 *
 * $Id: flimsy_dates.cxx,v 1.7 2002/02/08 16:03:31 esm Exp $
 *
 * Copyright  2002 Eduardo Santiago Muoz
 *
 * This file is part of Flimsy.
 *
 * Flimsy is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flimsy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flimsy; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "defs.h"
#include "flimsy_ui.h"

/*
** Hardcoded list of months
*/
const char *MoY[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
		      "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

const char *DoW[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

/*****************************************************************************\
|*                    Time-related functions                                 *|
\*****************************************************************************/
/*
** These two functions are used to compress showtimes (HH:MM) into bytes.
**
** Probably a waste of time... but for a long list of movies, the space
** savings may add up.  Not everyone lives in Los Alamos, where there
** are only a dozen theaters within a 100-mile radius.
*/

/**************\
**  encode_t  **  Encode a showtime in one byte.
\**************/
unsigned char encode_t(unsigned long t)
{
  /*
  ** If we break down the day into 5-minute intervals, we can almost
  ** scrunch up a day into one byte: 12 (60/5) intervals per hour,
  ** times 24 hours, is 288.  Not quite... it would have to be a
  ** big byte.
  **
  ** But we're so close... so let's assume that no movies are scheduled
  ** for 04:00 through 07:00 -- not an unreasonable assumption?  This
  ** gives us three hours, hence 12*21=252 which fits.
  **
  ** Finally, to accommodate the night owls, change the reference
  ** base, so the day starts at 7:00 AM and goes on to 04:00 AM.
  ** This then shows, e.g., 00:10 later than 23:45
  */
  if (t < 400) {
    t += 2400;
  }
  if (t <= 700) {
    t = 0;
  } else {
    t -= 700;
  }

  /*
  ** t is a stupidly-formatted time, "HHMM".  For instance, 1730 is 17:30,
  ** or 5:30 PM.
  **
  ** The upper half (t / 100) will thus be 0-23, and the lower half 0-59.
  ** The "12" in the equation below has nothing to do with 12 hours: it
  ** is the number of 5-minute intervals in an hour (60/5).  Remember,
  ** we've picked 5 minutes as our time quantum.  The RHS is obvious.
  **
  ** Finally, be sure to add 1 to the result, so our range is 1-253
  ** instead of 0-252.  This lets us use "0" as an error case, if
  ** we ever need to.  Maybe we'll use it for "movie starts this
  ** day, but we don't know the times".
  */
  return (t/100) * 12 + (t % 100) / 5 + 1;
}


/**************\
**  decode_t  **  Given an encoded (byte) time, returns "HH:MM" string
\**************/
const char *decode_t(unsigned char t_encoded)
{
  static char buf[10];
  unsigned short i = (t_encoded-1) * 5;	// given 5-minute period in the day
  unsigned short t = (i/60)*100 + (i%60);

  /*
  ** Above, we collapsed all times between 0400 and 0700.  Now we
  ** un-collapse.  This removes any possibility of distinguishing
  ** between 4:00 and 7:00.  But we don't care.
  */
  t += 700;
  t %= 2400;

  sprintf(buf,"%02d:%02d", t/100, t%100);

  return buf;
}


/*****************************************************************************\
|*                    Date-related functions                                 *|
\*****************************************************************************/

/*
**
*/
char const *flimsy_ui::friendly_date( void )
{
  // Determine the day delta
  int delta = 8;
  unsigned char todays_bit_copied = todays_bit;
  while (todays_bit_copied != 0) {
    delta--;
    todays_bit_copied >>= 1;
  }

  return friendly_date( delta );
}

char const *flimsy_ui::friendly_date( int delta )
{
  struct tm tm;
  static char buf[255];

  tm.tm_year = base_date.tm_year;
  tm.tm_mon  = base_date.tm_mon;
  tm.tm_mday = base_date.tm_mday + delta;

  // Query 12:12:12 local time, so we aren't caught by daylight-savings bugs
  tm.tm_hour = tm.tm_min = tm.tm_sec = 12;

  time_t t = mktime(&tm);
  struct tm *new_tm = localtime(&t);

  sprintf(buf, "%s %d %s", DoW[new_tm->tm_wday],
	  new_tm->tm_mday, MoY[new_tm->tm_mon]);

  return buf;
}

/*********************\
**  days_since_base  **  How many days old is the source input file?
\*********************/
int flimsy_ui::days_since_base( void )
{
  struct tm tm;

  // Get a tm struct for today.  memcpy() it, since localtime() clobbers.
  time_t       now = time(NULL);
  struct tm tm_now;
  memcpy(&tm_now, localtime(&now), sizeof(struct tm));

  /*
  ** Loop from 0 to 7 days in the future, comparing dates.  The
  ** complicated fill-in-tm, mktime(), localtime() gobbledygook
  ** is because mktime() can accept crap (e.g., "Jan 35"), but
  ** localtime() will return it normalized.
  */
  for (int i=0; i < 8; i++) {
    tm.tm_year = base_date.tm_year;
    tm.tm_mon  = base_date.tm_mon;
    tm.tm_mday = base_date.tm_mday + i;

    // Query 12:12:12 local time, so we aren't caught by daylight-savings bugs
    tm.tm_hour = tm.tm_min = tm.tm_sec = 12;

    time_t t = mktime(&tm);
    struct tm *tm_then = localtime(&t);

    // Compare.  Is this today?  Return the index.
    if (tm_then->tm_year == tm_now.tm_year &&
	tm_then->tm_mon  == tm_now.tm_mon  &&
	tm_then->tm_mday == tm_now.tm_mday) {
      return i;
    }
  }

  // Oh well.  File is too old.
  fl_alert("XML file is way too old, sorry!");

  // FIXME: Return 0, so user can still see *something*?  Or just die?
  return 0;
}


/********************\
**  date_change_cb  **  callback for the "Date" selection menu
\********************/
void flimsy_ui::date_change_cb(Fl_Widget *w, void *p)
{
  Fl_Menu_Button *x = (Fl_Menu_Button*)w;
  flimsy_ui *flimsy = (flimsy_ui*)p;
  int n = x->value();

  if (n == flimsy->days_since_base()) {
    flimsy->date_mb->label("Today");
  } else {
    flimsy->date_mb->label(flimsy->friendly_date(n));
  }
  flimsy->date_mb->redraw();

  flimsy->todays_bit = 1 << (7 - n);

  // FIXME: figure out a way to remember which page we're on, and what line #?
  flimsy->show_by_theater(0);
}


/**************************\
**  something_playing_on  **  Can we find any listings for this given day?
\**************************/
int flimsy_ui::something_playing_on( unsigned char mask )
{
  for (int listing=0; showtimes[listing]; listing++) {
    for (int t=0; t < showtimes[listing][2]; t++) {
      if (showtimes[listing][3+2*t] & mask) {
	return 1;
      }
    }
  }
  return 0;
}


/***********************\
**  fill_in_date_menu  **  Fills in the days of the week, and labels the Menu
\***********************/
void flimsy_ui::fill_in_date_menu( void )
{
  char buf[30];

  date_mb->clear();
  for (int i=0; i < 7; i++) {
    unsigned char mask = 1 << (7-i);
    int flag = FL_MENU_INACTIVE;	// Assume the menu item is inactive...

    // ...unless there are actually any listings for this day.
    if (something_playing_on(mask)) {
      flag = 0;
    }

    strcpy(buf, friendly_date(i));
    if (todays_bit == mask) {
      strcat(buf, " *** TODAY ***");
    }
    date_mb->add( buf, 0, date_change_cb, (void*)this, flag);
  }

  date_mb->label("Today");
  date_mb->redraw();
}
