#!/usr/bin/perl -w
#
# flimsy_get_listings - get movie listings, convert to Agenda format
#
# $Id: flimsy_get_listings.pl,v 1.15 2002/02/06 00:40:29 esm Exp $
#
(our $ME = $0) =~ s|.*/||;
(our $VERSION = '$Revision: 1.15 $ ') =~ tr/[0-9].//cd;

use strict;
use LWP;
use Tie::IxHash;		# Indexed hash: preserves order
use Time::localtime;		# OO interface to localtime

###############################################################################
# BEGIN user-customizable section

# My local area code.  It is stripped out from the Phone number shown in
# the listings.
our $LOCAL_AREA_CODE = 505;

# Name of output file.
our $Outfile = 'flimsy.xml';

# Yahoo URL, first argument is ZIP code, second is delta-T, or, number
# of days ahead.  If you change this to use another site, you will need
# to do some major surgery to the HTML::Filter code below!
our $URL = "http://movies.yahoo.com/showtimes/showtimes.html?z=%d&dt=%d";

# Number of days to prefetch, starting with today
our $ndays = 7;

# Days of the week, in two-letter form
our @DoW = qw(Su Mo Tu We Th Fr Sa);

# END   user-customizable section
###############################################################################

###############################################################################
# BEGIN boilerplate args checking, usage messages

sub usage($) {
  my ($exit_code) = @_;
  my $STREAM = ($exit_code == 0 ? *STDOUT : *STDERR);
  if ($exit_code != 0) {
    print  $STREAM "Try `$ME -help' for more information.\n";
  } else {
    print  $STREAM <<EOH;
Usage: $ME ZIPCODE [ZIPCODE...]

$ME queries Yahoo for movie showtimes, and
writes out an XML file suitable for use by the "flimsy" tool.

OPTIONS:

  --ndays=#    fetch listings for today through # days ahead (default: $ndays)

  --help       display this message
  --version    display program name and version
EOH
  }

  exit $exit_code;
}

# Command-line options
my $debug   = 0;
use Getopt::Long;
GetOptions( 'debug!'     => \$debug,
	    'ndays=i'    => \$ndays,
	    help         => sub { usage 0 },
	    version      => sub { print "$ME version $VERSION\n"; exit 0 },
) or usage(1);

@ARGV
  or (warn "$ME: Missing ZIP code\n"),usage 1;

# END   boilerplate args checking, usage messages
###############################################################################

###############################################################################
# BEGIN ListingsFilter, a HTML::Filter package which does the parsing

package ListingsFilter;

use strict;
use warnings;
use HTML::Filter;
our @ISA = qw(HTML::Filter);


sub unentify {
  my $s = shift;

  $s =~ s/\&nbsp;/ /g;
  $s =~ s/\&amp;/ /g;

  $s;
}

our @Order = qw(Name Phone Address);

sub start {
  my $self = shift;
  my $tag = shift;

  if ($tag eq 'a' && @_) {
    my $att = shift;
    ref $att eq 'HASH'
      or die "Eek!  Not a hashref!";
    if (exists $att->{name} && $att->{name} =~ /^T/) {
      $self->{in_theater} = 1;
      $self->{td_num} = 0;
      $self->{movie_num} = -1;
      if (keys %{$self->{current_record}} > 1) {
	push @{$self->{all}}, $self->{current_record};
      }
      $self->{current_record} = { movies => [] };
    } elsif (exists $att->{href} && $att->{href} =~ m!/shop\?d=hv!) {
      $self->{movie_num}++;
    }
  }

  if ($tag eq 'small') {
    $self->{is_small} = 1;
  }

  $tag =~ /^(table|tr|td)$/
    or return;

  if ($tag eq 'td') {
    $self->{in_td} = 1;
    $self->{td_num}++;
  } elsif ($tag eq 'tr') {
    $self->{in_td} = 0;
  }
}

sub end {
  my $self = shift;
  my $tag = shift;

  if ($tag eq 'small') {
    $self->{is_small} = 0;
  }

  if ($tag eq 'td' || $tag eq 'tr' || $tag eq 'table') {
    $self->{in_td} = 0;
#    if ($tag eq 'table') {
#      $self->{in_theater} = 0;
#    }
  }
}

sub text {
  my $self = shift;

  $self->{in_theater} or return;

  my $text = shift;
  $text =~ /^\s*$/m and return;

  $self->{in_td} or return;

  $self->{is_small} and return;

  if ($self->{td_num} <= $#Order) {
    $self->{current_record}->{$Order[$self->{td_num}]} .= unentify($text);
  } elsif ($self->{movie_num} >= 0) {
#    use Data::Dumper; print Dumper($self->{current_record});
    $self->{current_record}->{movies}->[$self->{movie_num}] .= "--" . unentify($text);
  }

  # Debugging printf: shows the text we've seen
  printf "---- %s -- %s\n", $Order[$self->{td_num}] || "?", $text if $debug&16;
}

# We do no output
sub output {}

# END   ListingsFilter, a HTML::Filter package which does the parsing
###############################################################################

package main;

###############################################################################
# BEGIN helper functions


#############
#  bitmask  #  Given a days-in-advance number (0-7), returns the bitmask for it
#############
#
# Note that this is reversed from what you would expect: instead of LSB first,
# we use the more readable (to Western folk) MSB first.  Thus 0 is 10000000,
# 1 is 01000000, and so on.
sub bitmask($) {
  return 1 << (7 - shift);
}



############
#  entify  #  Convert to entities
############
sub entify($) {
  my $s = shift;

  $s =~ s/\&(?!\w{2,4}\;)/&amp;/g;
  $s =~ s/\"/&quot;/g;
  $s =~ s/\'/&apos;/g;

  $s;
}


################
#  add_to_all  #  Given the results of our ListingsFilter, add to %all
################
sub add_to_all($$@) {
  my $dt   = shift;
  my $href = shift;

  #
  # The input array, generated by the ListingsFilter above, is an array
  # of hashes, each of which look something like:
  #
  #    {
  #      'Name'    => 'Trans-Lux Dream Catcher'
  #      'Address' => '33771 South US Highway 285, Espanola, NM 87532',
  #      'Phone'   => 'Phone: (505)753-0087',
  #      'movies'  => [
  #        '--Black Hawk Down--(12:45 PM), (3:45), 6:45, 9:40  ',
  #        '--Kung Pow: Enter the Fist--(4:10 PM), 7:10  ',
  #        '--Lord of the Rings: The Fellowship of the Ring, The--(12:20 PM), (3:50), 7:30  ',
  #        '--Mothman Prophecies, The--(12:50 PM), (4:00), 7:05, 9:35  ',
  #        '--Slackers--(1:05 PM), (4:05), 7:15, 9:30  ',
  #        '--Snow Dogs--(1:00 PM), (4:15), 7:00, 9:20  ',
  #        '--Vanilla Sky--(12:55 PM), 9:25  -- '
  #      ],
  #    },
  #
  # The first time we see a theater (usually on $dt=0), we initialize
  # the Name/Address/Phone bits in the %all hash.  Then, and on subsequent
  # times, we process all the movies and add showtimes/days.
  #
  foreach my $theater (@_) {
    # First time we see this theater?  Create a new Tie::IxHash entry for it
    # FIXME: Is it worth bothering to check for changes in phone/address?
    my $theater_name = $theater->{Name};
    if (!exists $href->{$theater_name}) {
      $href->{$theater_name} =
	{
	 Phone    => $theater->{Phone},
	 Address  => $theater->{Address},
	 listings => {},
	};

      # Some minor tweaking: remove the superfluous "Phone:" label,
      # and strip away our locla area code.
      $href->{$theater_name}->{Phone} =~ s/^Phone:\s+//;
      $href->{$theater_name}->{Phone} =~ s/^\s*\($LOCAL_AREA_CODE\)\s*//;

      # The 'listings' hash contains a series of movie names, and their
      # times.  Make it an indexed hash, just in case it's important to
      # preserve the order in which each theater shows its listings.
      tie %{$href->{$theater_name}->{listings}}, 'Tie::IxHash'
	or die "$ME: Unable to IxHash $theater_name: $!\n";
    }

    #
    # Process each movie for this theater (see above for a sample of what
    # our input looks like).  We extract the movie name, then for each
    # listed time, set today's bit.
    #
    foreach my $movie (@{$theater->{movies}}) {
      my $pm = 0;

      # Extract the movie name.  This is the first string between '--'s
      $movie =~ /^\-\-(.*?)\-\-(.*)$/
	or die "$ME: Internal error: cannot grok movie name '$movie'\n";
      my ($name, $times) = ($1, $2);
      $name =~ s/\"/\\\"/g;

      # First time we see this movie?  Create a new hash, key=time, val=mask
      if (!exists $href->{$theater_name}->{listings}->{$name}) {
	$href->{$theater_name}->{listings}->{$name} = {};
      }

      # KLUDGE: Treat "Starts on Tuesday, Feb 5' as just one string
      $times =~ s/(Starts on \S+),(\s+\w+\s+\d+)/$1$2/g;
      $times =~ s/Buy Tickets://;

      foreach my $time (split(',',$times)) {
	$time =~ s/^\s*--//;
	$time =~ s/--\s*$//;		# Trim whitespace, and any '--'s

	# FIXME: we should probably indicate this somehow, so PDA
	# can distinguish between "not playing" and "not known"...?
	$time =~ /^Theater has not yet announced times/
	  and next;

	#                     1   1 2   2     3     3         4     4
	$time =~ /^\s*\(?\s*-*(\d+):(\d+)-*\s*(AM|PM)?-*\s*\)?(AM|PM)?-*\s*$/
	  or (warn "$ME: $theater_name: $name: Cannot grok time '$time'\n"),next;
	my ($hh, $mm, $ampm) = ($1, $2, $3 || $4);
	if (defined $ampm) {
	  $pm = ($ampm eq 'PM');
	  $hh %= 12;		# Otherwise, 12PM becomes midnight
	}

	# Simplify the time: make it just 'HHMM', for ease of parsing
	# in the C code.
	my $time = sprintf("%02d%02d", $hh+(12 * $pm), $mm);

	# Indicate that this showtime is valid on today's date.
	$href->{$theater_name}->{listings}->{$name}->{$time} |= bitmask($dt);
      }
    }
  }
}

# END helper functions
###############################################################################

############################## CODE BEGINS HERE ###############################


my $ua = new LWP::UserAgent;

my %all;
tie %all, 'Tie::IxHash'
  or die "$ME: Unable to make \%all a Tie::IxHash: $!\n";

foreach my $zip (@ARGV) {
  my $txt;

  $zip =~ /^(\d{5})$/
    or die "$ME: Illegal ZIP code '$zip'\n";
  $zip = $1;

  # Start today, but use our crystal ball to peer ahead at the future
  for (my $t=0; $t < $ndays; $t++) {
    my $cachefile = ".#$ME.cache.$zip.$t";
    my $txt;

    # For debugging: keep a cache file, to avoid lots of Yahoo queries
    if (-f $cachefile) {
      print STDERR "[reading from $cachefile]\n";
      # FIXME: Don't read if cachefile is > 1 day old??
      open IN, '<', $cachefile or die "$ME: Open( $cachefile ): $!\n";
      local $/ = undef;
      $txt = <IN>;
      close IN;
    } else {
      # No cache file: make HTTP query to remote server
      my $request = HTTP::Request->new (GET => sprintf($URL, $zip, $t));
      my $response = $ua->request($request);

      $response->is_success
	or die "$ME: GET failed: " . $response->error_as_HTML;

      $txt= $response->content;

      # Write out cache file?
      if ($debug) {
	unlink $cachefile;
	open OUT, '>', $cachefile
	  or die "$ME: Error creating $cachefile: $!\n";
	print OUT $txt;
	close OUT
	  or die "$ME: Error writing $cachefile: $!\n";
      }
    }

    # Parse the returned results.  ListingsFilter, above, does all the work.
    my $f = new ListingsFilter;
    $f->parse($txt);

    # Add it to our hash-of-all
    add_to_all $t, \%all, @{$f->{all}};
  }
}

#
# Make one pass over everything, so we can get movie names.  This
# gives us the movie key IDs.
#
my %mlist;
foreach my $t (keys %all) {
  foreach my $m (keys %{$all{$t}->{listings}}) {
    $mlist{$m}++;
  }
}


# We need to indicate, in the XML file, what the starting date is.
my $today = localtime;

#
# Start writing the XML file.  As always, we write to a temporary
# file, then if all works fine, move that into place.
#
my $tmp = "$Outfile.TMP.$$";
open OUT, '>', $tmp
  or die "$ME: Error creating $tmp: $!\n";
print  OUT "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n";
printf OUT "<movies base=\"%d-%02d-%02d\">\n", localtime->year+1900,
  localtime->mon+1, localtime->mday;

#
# First section, lists the movies in order and binds them to their key:
#
#      <movienames>
#        <bind key="0" name="Amelie" />
#        <bind key="1" name="Beautiful Mind, A" />
#        ....
#      </movienames>
#
my $i = 0;
printf OUT "  <movienames>\n";
foreach my $m (sort keys %mlist) {
  $mlist{$m} = $i;			# mlist now holds movie key, not count!

  # Emit movie key and name
  printf OUT "    <bind key=\"%d\" name=\"%s\" />\n", $i, entify($m);

  # Bump up counter.  This is done here, instead of inline above, just in
  # case we ever muck with the sequence above.  We don't want to lose this.
  ++$i;
}
printf OUT "  </movienames>\n\n";


#
# Now we get to the theaters.  This is a section containing zero or
# more <movie> entries
#
#     <theater name="Trans-Lux Dream Catcher" address="..." phone="...">
#       <movie key="3">     <!-- Black Hawk Down -->
#         <showing time="1245" days="10000000" />
#         <showing time="1545" days="11111000" />
#         <showing time="1845" days="11111000" />
#         <showing time="2140" days="11111000" />
#       </movie>
#       ....
#     </theater>
#
foreach my $t (keys %all) {
  # Theater name, address, and phone number
  printf OUT "  <theater name=\"%s\"", entify($t);
  foreach my $key (qw(Address Phone)) {
    (my $val = $all{$t}->{$key}) =~ s/\"/\\\"/g;
    printf OUT " %s=\"%s\"", lc $key, entify($val);
  }
  print  OUT ">\n";

  # List all the movies playing at this theater.  This will be the whole
  # set of movies found for the next $ndays days.
  foreach my $m (keys %{$all{$t}->{listings}}) {
    exists $mlist{$m}
      or die "$ME: Internal error: no key for movie '$m'\n";
    printf OUT "    <movie key=\"%d\">\t<!-- %s -->\n", $mlist{$m}, $m;

    # For each movie, list its showtimes.  Again, this will be the
    # entire list of unique showtimes we've encountered.  The "days"
    # bitmask takes care of identifying which are valid on what days.
    # For instance:
    #
    #      <movie key="12">    <!-- Kiss Me Deadly -->
    #      <showing time="1915" days="00010000" />   <!--       We         -->
    #      <showing time="1930" days="00101000" />   <!--     Tu  Th       -->
    #      </movie>
    #
    # ...indicates that the 19:30 showing is valid on Tuesday and Thursday,
    # but the movie plays at 19:15 on Wednesday.
    #
    foreach my $time (sort { $a <=> $b } keys %{$all{$t}->{listings}->{$m}}) {
      my $mask = $all{$t}->{listings}->{$m}->{$time};
      printf OUT "      <showing time=\"%s\" days=\"%08b\" />", $time, $mask;

      # For human-friendliness: emit comment, showing days of week
      print OUT "\t<!-- ";
      for (my $dt=0; $dt < 8; $dt++) {
	if ($mask & bitmask($dt)) {
	  print OUT $DoW[($today->wday + $dt) % @DoW];
	} else {
	  print OUT "  ";
	}
      }
      print OUT " -->\n";
    }
    print  OUT "    </movie>\n";
  }
  print  OUT "  </theater>\t<!-- $t -->\n\n";
}

print OUT "</movies>\n";

#
# Done!  We've finished writing out the XML file.  Now check for any
# errors when closing.  If all is OK, rename the file into place.  If
# an older copy exists, keep a backup of it.  Just in case.
#
close OUT
  or die "$ME: Error writing $tmp: $!\n";

chmod 0444, $tmp;
link           $Outfile => "$Outfile.BAK"		if -f $Outfile;
rename $tmp => $Outfile
  or die "$ME: Coult not rename $tmp: $!\n";

exit 0;
