/*
 * flimsy_xml.cxx - Parses the input XML file, stores it in compact form
 *
 * $Id: flimsy_xml.cxx,v 1.19 2002/02/04 15:04:05 esm Exp $
 *
 * Copyright  2002 Eduardo Santiago Muoz
 *
 * This file is part of Flimsy.
 *
 * Flimsy is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flimsy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flimsy; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "defs.h"
#include "flimsy_ui.h"

#include <xmlparse.h>

#define	BUFSIZE	1024
static char buf[BUFSIZE];


/*****************************************************************************\
|*                          XML Handlers                                     *|
\*****************************************************************************/

/*
** Globals, ick.  But it's not worth the effort to do them better.
*/
static int current_theater;			// Current theater ID
static unsigned char tmp_showtimes[255];	// Temp holding buffer

/********************\
**  find_attribute  **  Find "key" in given list of key/value pairs
\********************/
static char const *find_attribute(const char **attr, const char *key)
{
  /*
  ** attr is an array of pointers to strings; it must be an even number
  ** of elements, since each first is a key name, and each second is
  ** its value.  Think Perl hashes: $x{key} = 'value', ...
  */
  for (int i=0; attr[i]; i += 2) {
    if (!strcmp(attr[i], key)) {
      return attr[i+1];
    }
  }

  // Uh-oh.  This is not supposed to happen... it probably means
  // someone munged the XML file by hand, or with bad code.
  char buf[1024];
  sprintf(buf, "Key '%s' not found!  Probably a bad XML file", key);
  fl_alert(buf);
  exit(1);
}




/******************\
**  handle_start  **  handle xml start tags, e.g., <foo>
*******************
**
** This does most of the work, because Ed was lazy in his definition
** of the XML file, and put most of the details as attributes instead
** of CDATA.
**
** Here's a very brief example of what the XML data file looks like:
**
**    <movies>
**      <movienames>
**        <bind key="1" name="Amelie" />
**        <bind key="2" name="Beautiful Mind, A" />
**      </movienames>
**      <theater name="My Theater" address="123 45th St" phone="555-1393">
**        <movie key="2">
**          <showing time="1920" days="11100000" />
**          <showing time="2130" days="11111000" />
**        </movie>
**      </theater>
**
*/
static void handle_start(void *data, const char *el, const char **attr)
{
  flimsy_ui* flimsy = (flimsy_ui*)data;
  int i;

  /*
  ** <movies base="YYYY-MM-DD">
  **
  ** This is what starts it all off.  We need to parse the date,
  ** so we know how to handle the "days" bitmask for each showtime.
  */
  if (!strcmp(el, "movies")) {
    char *base = (char*)find_attribute(attr, "base");

    flimsy->base_date.tm_year = strtol(base,   &base, 10) - 1900;
    flimsy->base_date.tm_mon  = strtol(base+1, &base, 10) -    1;
    flimsy->base_date.tm_mday = strtol(base+1, &base, 10);

    return;
  }

  /*
  ** "bind" is used to bind a movie name to a numeric index:
  **
  **    <bind key="16" name="Lord of the Rings: The Fellowship ..." />
  */
  if (!strcmp(el,"bind")) {
    long key = strtol(find_attribute(attr,"key"), NULL, 10);
    char const *name = find_attribute(attr,"name");

    if (key < 0  ||  255 < key) {
      fl_alert("Invalid key '%ld' for movie '%s'", key, name);
      exit(1);
    }
    if (flimsy->movie_name[key]) {
      fl_alert("Duplicate key '%ld' ('%s', '%s')", key,
	       flimsy->movie_name[key], name);
      exit(1);
    }

    // Allocate space for this movie name, and make a permanent copy.
    flimsy->movie_name[key] = (char*)malloc(strlen(name)+1);
    strcpy(flimsy->movie_name[key], name);
  }


  /*
  ** <theater name="My Theater" address="..." phone="...">
  **
  ** There is no key, because we just take that from the ordinal
  ** index, that is, its position in the XML file.  The first theater
  ** we see is 0, then 1, 2, etc.
  **
  ** We preserve this index in the global current_theater var, because
  ** it will need to be specified as part of each movie showing.  We
  ** could as easily determine it from the last non-null theater_info[],
  ** but I don't feel like taking the time just to remove a global.
  */
  if (!strcmp(el,"theater")) {
    // Find the first empty slot in the theater_info array.
    for (i=0; i < MAX_THEATERS; i++) {
      if (flimsy->theater_info[i].name == NULL) {
	char const *name = find_attribute(attr,"name");
	flimsy->theater_info[i].name = (char*)malloc(strlen(name)+1);
	strcpy(flimsy->theater_info[i].name, name);

	current_theater = i;

	// Address and phone are optional
	for (int j=0; attr[j]; j += 2) {
	  if (!strcmp(attr[j], "address")) {
	    flimsy->theater_info[i].address = (char*)malloc(strlen(attr[j+1])+1);
	    strcpy(flimsy->theater_info[i].address, attr[j+1]);
	  }
	  if (!strcmp(attr[j], "phone")) {
	    flimsy->theater_info[i].phone = (char*)malloc(strlen(attr[j+1])+1);
	    strcpy(flimsy->theater_info[i].phone, attr[j+1]);
	  }
	}

	// FIXME: include phone, address
	return;
      }
    }

    // Uh-oh!  Did not find a place for this theater!  Someone is
    // going to have to move to a smaller city.
    fl_alert("Out of space for theaters!");
    // Probably don't need to exit, but we should reset current_theater
    // and make sure we don't list any showtimes for this one.
    current_theater = -1;
  }

  /*
  ** "movie" is used inside a "theater" section to indicate that
  ** the movie given by numeric key 'key' is playing in that theater.
  ** Inside the movie section is a series of one or more times.
  **
  **    <movie key="16">
  **
  ** When we find this, we immediately store the current theater ID
  ** in our tmp_showtimes array, followed by the given movie key,
  ** followed immediately by a count, "0", of showtimes.  These
  ** will be filled in by subsequent <showing> tags.
  */
  if (!strcmp(el,"movie")) {
    long movie_key = strtol(find_attribute(attr,"key"), NULL, 10);

    if (movie_key < 0 || 255 < movie_key) {
      fl_alert("Invalid movie key '%ld'\n(in theater '%s')",
	       movie_key, flimsy->theater_info[current_theater].name);
      exit(1);
    }

    tmp_showtimes[0] = current_theater;
    tmp_showtimes[1] = (unsigned char)movie_key;
    tmp_showtimes[2] = 0;
  }

  /*
  ** One (or probably many more) of these are expected to follow
  ** immediately after a <movie> tag.  It looks like:
  **
  **      <showing time="1605" days="1000000000" />
  **
  ** time is simply HHMM, which we encode into a byte (see above)
  **
  ** days is tricky.  It's a bit mask of days for which this time
  ** is valid.  Each bit is one day, starting with the day on which
  ** the XML file was generated.  Just to make things harder for
  ** me (the programmer), but easier for humans to read, the mask
  ** reads MSB-first, that is, left-to-right.  Unless your native
  ** language is Hebrew, that probably makes more sense to you,
  ** and there's some value to being able to look at the XML file
  ** to check its validity.
  */
  if (!strcmp(el,"showing")) {
    unsigned char mask = strtol(find_attribute(attr,"days"), NULL, 2);
    unsigned long t    = strtol(find_attribute(attr,"time"), NULL, 10);

    tmp_showtimes[3 + 2*tmp_showtimes[2]]   = mask;
    tmp_showtimes[3 + 2*tmp_showtimes[2]+1] = encode_t(t);

    // Increment the count.  This is a count of SHOWINGS, not bytes!
    tmp_showtimes[2]++;
  }
}


/****************\
**  handle_end  **  Handle </tag> corresponding to the above <tag>
\****************/
static void handle_end(void *data, const char *el)
{
  flimsy_ui* flimsy = (flimsy_ui*)data;

  /*
  ** The only one we care about is </movie>.
  **
  ** When we see it, we allocate space for a new "showtimes" listing,
  ** and copy from our tmp_showtimes[]
  */
  if (!strcmp(el,"movie")) {
    for (int i=0; i < MAX_SHOWTIMES; i++) {
      if (flimsy->showtimes[i] == NULL) {
	flimsy->showtimes[i] = (unsigned char*)malloc(tmp_showtimes[2]*2+3);
	memcpy(flimsy->showtimes[i], tmp_showtimes, tmp_showtimes[2]*2+3);
	return;
      }
    }
  }
}

/** END OF XML HANDLERS **/


/*****************************************************************************\
|*                          The Main Code                                    *|
\*****************************************************************************/

/****************\
**  parse_data  **  Parses the XML file, fills in the movie/theater info
\****************/
void flimsy_ui::parse_data(const char *filename)
{
  int done;
  FILE* fp;
  XML_Parser p = XML_ParserCreate(NULL);

  // Initialize.  In case we've been called before, clear out any
  // allocate memory we may have used.
  for (int i=0; i < 255; i++) {
    if (theater_info[i].name) {
      free(theater_info[i].name);
      theater_info[i].name = NULL;
    }
    if (theater_info[i].address) {
      free(theater_info[i].address);
      theater_info[i].address = NULL;
    }
    if (theater_info[i].phone) {
      free(theater_info[i].phone);
      theater_info[i].phone = NULL;
    }
    if (movie_name[i]) {
      free(movie_name[i]);
      movie_name[i] = NULL;
    }
  }

  /*
  ** Expat setup: use handle_start() and handle_end() to handle <tag>
  ** and </tag> respectively, and pass "this" to each one as the UserData.
  */
  XML_SetElementHandler(p, handle_start, handle_end);
  XML_SetUserData(p, (void*)this);

  // Open the XML file for reading.  If this doesn't work, there's just
  // not much we can do other than bail out.
  fp = fopen(filename, "r");
  if (fp == NULL) {
    fl_alert("Unable to open %s", filename);
    exit(1);
  }

  do {
    size_t len = fread(buf, 1, sizeof(buf), fp);
    done = len < sizeof(buf);

    if (!XML_Parse(p, buf, len, done)) {
      fprintf(stderr, "%s at line %d\n", XML_ErrorString(XML_GetErrorCode(p)),
	                                 XML_GetCurrentLineNumber(p));
    }
  } while (!done);

  fclose(fp);
  XML_ParserFree(p);
}
